"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getMaxAttempts = exports.nackInbox = exports.ackInbox = exports.verifyInbox = exports.initializeInboxMessageStorage = void 0;
const pg_1 = require("pg");
const error_1 = require("../common/error");
const utils_1 = require("../common/utils");
/**
 * Initialize the inbox message storage to store incoming messages in the inbox table.
 * @param config The configuration object that defines the values on how to connect to the database and general settings.
 * @param logger A logger instance for logging trace up to error logs
 * @returns The function to store the inbox message data to the database and the shutdown action.
 */
const initializeInboxMessageStorage = (config, logger) => {
    const pool = new pg_1.Pool(config.pgConfig);
    pool.on('error', (err) => {
        logger.error(err, 'PostgreSQL pool error');
    });
    /**
     * The function to store the inbox message data to the database.
     * @param message The received message that should be stored as inbox message
     * @throws Error if the inbox message could not be stored
     */
    return [
        (message) => __awaiter(void 0, void 0, void 0, function* () {
            try {
                yield (0, utils_1.executeTransaction)(pool, (client) => __awaiter(void 0, void 0, void 0, function* () {
                    yield insertInbox(message, client, config, logger);
                }), logger);
            }
            catch (err) {
                logger.error(Object.assign(Object.assign({}, message), { err }), 'Could not store the inbox message');
                throw new error_1.MessageError(`Could not store the inbox message with id ${message.id}`, message, (0, error_1.ensureError)(err));
            }
        }),
        () => __awaiter(void 0, void 0, void 0, function* () {
            pool.removeAllListeners();
            try {
                yield pool.end();
            }
            catch (e) {
                logger.error(e, 'Inbox message storage shutdown error');
            }
        }),
    ];
};
exports.initializeInboxMessageStorage = initializeInboxMessageStorage;
/**
 * Make sure the inbox item was not and is not currently being worked on. And
 * set the actual attempts and processed_at values for the WAL message.
 * As the WAL inbox service does not run in the same transaction as the message
 * handler code there is a small chance that the handler code succeeds but the
 * WAL inbox message was not acknowledged. This takes care of such cases.
 * @param message The inbox message to check and later process.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 * @throws Throws an error if the database row does not exist, is locked by some other transaction, or was already processed.
 */
const verifyInbox = (message, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    // Get the inbox data and lock it for updates. Use NOWAIT to immediately fail if another process is locking it.
    const inboxResult = yield client.query(
    /* sql*/ `SELECT processed_at, attempts FROM ${settings.dbSchema}.${settings.dbTable} WHERE id = $1 FOR UPDATE NOWAIT`, [message.id]);
    if (inboxResult.rowCount === 0) {
        return 'INBOX_MESSAGE_NOT_FOUND';
    }
    const dbItem = inboxResult.rows[0];
    if (dbItem.processed_at) {
        return 'ALREADY_PROCESSED';
    }
    if (dbItem.attempts >= (0, exports.getMaxAttempts)(settings.maxAttempts)) {
        return 'MAX_ATTEMPTS_EXCEEDED';
    }
    // assign the actual attempts/processed_at to the WAL message
    message.attempts = dbItem.attempts;
    message.processedAt = dbItem.processed_at;
    return true;
});
exports.verifyInbox = verifyInbox;
/**
 * Acknowledge that the inbox message was handled.
 * @param message The inbox message to acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 */
const ackInbox = ({ id }, client, { settings }) => __awaiter(void 0, void 0, void 0, function* () {
    yield client.query(
    /* sql*/ `UPDATE ${settings.dbSchema}.${settings.dbTable} SET processed_at = $1, attempts = attempts + 1 WHERE id = $2`, [new Date().toISOString(), id]);
});
exports.ackInbox = ackInbox;
/**
 * Reject the message and do not acknowledge it. Check the return value if the
 * WAL message should be acknowledged or not as the logic uses retries.
 * @param message The inbox message to NOT acknowledge.
 * @param client The database client. Must be part of the transaction where the message handling changes are done.
 * @param config The configuration settings that defines inbox database schema.
 * @param attempts Optionally set the number of attempts to a specific value. Especially relevant for non-transient errors to directly set it to the maximum attempts value.
 */
const nackInbox = ({ id }, client, { settings }, attempts) => __awaiter(void 0, void 0, void 0, function* () {
    if (attempts) {
        yield client.query(
        /* sql*/ `
      UPDATE ${settings.dbSchema}.${settings.dbTable} SET attempts = $1 WHERE id = $2;`, [attempts, id]);
    }
    else {
        yield client.query(
        /* sql*/ `
    UPDATE ${settings.dbSchema}.${settings.dbTable} SET attempts = attempts + 1 WHERE id = $1;`, [id]);
    }
});
exports.nackInbox = nackInbox;
/** Gets the maximum attempts to process a message. Defaults to 5 if not configured. */
const getMaxAttempts = (maxAttempts) => maxAttempts !== null && maxAttempts !== void 0 ? maxAttempts : 5;
exports.getMaxAttempts = getMaxAttempts;
const insertInbox = (message, dbClient, { settings }, logger) => __awaiter(void 0, void 0, void 0, function* () {
    const { id, aggregateType, aggregateId, messageType, payload, metadata, createdAt, } = message;
    const inboxResult = yield dbClient.query(
    /* sql*/ `
    INSERT INTO ${settings.dbSchema}.${settings.dbTable}
      (id, aggregate_type, aggregate_id, message_type, payload, metadata, created_at)
      VALUES ($1, $2, $3, $4, $5, $6, $7)
      ON CONFLICT (id) DO NOTHING`, [id, aggregateType, aggregateId, messageType, payload, metadata, createdAt]);
    if (!inboxResult.rowCount || inboxResult.rowCount < 1) {
        logger.warn(message, `The message with id ${id} already existed`);
    }
});
//# sourceMappingURL=inbox.js.map